from copy import deepcopy
import alpaqa as pa
import numpy as np
import concurrent.futures
import pytest
import os


@pytest.mark.skipif(not pa.with_casadi, reason="requires CasADi")
def test_alm_threaded():
    valgrind = 'valgrind' in os.getenv('LD_PRELOAD', '')

    pp = pa.PANOCParams(max_no_progress=100, max_iter=100)
    lbfgs = pa.LBFGSDirection()
    panoc = pa.PANOCSolver(pp, lbfgs)
    alm_params = pa.ALMParams(tolerance=1e-200, dual_tolerance=1e-200, max_iter=200, print_interval=0)
    solver = pa.ALMSolver(alm_params, panoc)

    import casadi as cs

    n = 2
    m = 2
    x = cs.SX.sym("x", n)

    Q = np.array([[1.5, 0.5], [0.5, 1.5]])
    f = 0.5 * x.T @ Q @ x
    g = x
    D = [-np.inf, 0.5], [+np.inf, +np.inf]

    name = "testproblem"
    p = pa.minimize(f, x).subject_to(g, D).with_name(name).compile()
    p = pa.Problem(p)

    def good_experiment():
        _, _, stats = deepcopy(solver)(deepcopy(p), asynchronous=True)
        return stats

    def bad_experiment1():
        _, _, stats = solver(deepcopy(p), asynchronous=True)
        return stats

    def bad_experiment2():
        _, _, stats = deepcopy(solver)(p, asynchronous=True)
        return stats

    def run(experiment):
        N = 4 if valgrind else 200
        with concurrent.futures.ThreadPoolExecutor(max_workers=os.cpu_count()) as pool:
            futures = (pool.submit(experiment) for _ in range(N))
            for future in concurrent.futures.as_completed(futures):
                stats = future.result()
                assert stats["status"] == pa.SolverStatus.MaxIter

    run(good_experiment)
    if not valgrind:
        with pytest.raises(
            RuntimeError, match=r"^Same instance of ALMSolver<PANOCSolver<LBFGS"
        ) as e:
            run(bad_experiment1)
        print(e.value)
        with pytest.raises(
            RuntimeError, match=r"^Same instance of testproblem used in multiple threads"
        ) as e:
            run(bad_experiment2)
        print(e.value)


if __name__ == "__main__":
    test_alm_threaded()
    print("done.")



import alpaqa
import numpy as np

class MyProblem:
    def __init__(self):
        self.num_variables = 3 # Number of variables
        self.num_constraints = 2 # Number of constraints
    def eval_projecting_difference_constraints(self, z: np.ndarray, e: np.ndarray) -> None: ...
    def eval_projection_multipliers(self, y: np.ndarray, M: float) -> None: ...
    def eval_proximal_gradient_step(self, γ: float, x: np.ndarray, grad_ψ: np.ndarray, x_hat: np.ndarray, p: np.ndarray) -> float: ...
    def eval_inactive_indices_res_lna(self, γ: float, x: np.ndarray, grad_ψ: np.ndarray, J: np.ndarray) -> int: ...
    def eval_objective(self, x: np.ndarray) -> float: ...
    def eval_objective_gradient(self, x: np.ndarray, grad_fx: np.ndarray) -> None: ...
    def eval_constraints(self, x: np.ndarray, gx: np.ndarray) -> None: ...
    def eval_constraints_gradient_product(self, x: np.ndarray, y: np.ndarray, grad_gxy: np.ndarray) -> None: ...
    def eval_grad_gi(self, x: np.ndarray, i: int, grad_gi: np.ndarray) -> None: ...
    def eval_lagrangian_hessian_product(self, x: np.ndarray, y: np.ndarray, scale: float, v: np.ndarray, Hv: np.ndarray) -> None: ...
    def eval_augmented_lagrangian_hessian_product(self, x: np.ndarray, y: np.ndarray, Σ: np.ndarray, scale: float, v: np.ndarray, Hv: np.ndarray) -> None: ...
    def eval_objective_and_gradient(self, x: np.ndarray, grad_fx: np.ndarray) -> float: ...
    def eval_objective_and_constraints(self, x: np.ndarray, g: np.ndarray) -> float: ...
    def eval_objective_gradient_and_constraints_gradient_product(self, x: np.ndarray, y: np.ndarray, grad_f: np.ndarray, grad_gxy: np.ndarray) -> None: ...
    def eval_lagrangian_gradient(self, x: np.ndarray, y: np.ndarray, grad_L: np.ndarray, work_n: np.ndarray) -> None: ...
    def eval_augmented_lagrangian(self, x: np.ndarray, y: np.ndarray, Σ: np.ndarray, ŷ: np.ndarray) -> float: ...
    def eval_augmented_lagrangian_gradient(self, x: np.ndarray, y: np.ndarray, Σ: np.ndarray, grad_ψ: np.ndarray, work_n: np.ndarray, work_m: np.ndarray) -> None: ...
    def eval_augmented_lagrangian_and_gradient(self, x: np.ndarray, y: np.ndarray, Σ: np.ndarray, grad_ψ: np.ndarray, work_n: np.ndarray, work_m: np.ndarray) -> float: ...
    def get_variable_bounds(self) -> alpaqa.Box: ...
    def get_general_bounds(self) -> alpaqa.Box: ...
    def check(self): ...

def test_problem_obj():
    problem = alpaqa.Problem(MyProblem())

    assert problem.num_variables == 3
    assert problem.num_constraints == 2
    assert problem.provides_eval_inactive_indices_res_lna()
    assert not problem.provides_eval_constraints_jacobian()
    assert not problem.provides_get_constraints_jacobian_sparsity()
    assert problem.provides_eval_grad_gi()
    assert problem.provides_eval_lagrangian_hessian_product()
    assert not problem.provides_eval_lagrangian_hessian()
    assert not problem.provides_get_lagrangian_hessian_sparsity()
    assert problem.provides_eval_augmented_lagrangian_hessian_product()
    assert not problem.provides_eval_augmented_lagrangian_hessian()
    assert not problem.provides_get_augmented_lagrangian_hessian_sparsity()
    assert problem.provides_eval_objective_and_gradient()
    assert problem.provides_eval_objective_and_constraints()
    assert problem.provides_eval_objective_gradient_and_constraints_gradient_product()
    assert problem.provides_eval_lagrangian_gradient()
    assert problem.provides_eval_augmented_lagrangian()
    assert problem.provides_eval_augmented_lagrangian_gradient()
    assert problem.provides_eval_augmented_lagrangian_and_gradient()
    assert problem.provides_get_variable_bounds()
    assert problem.provides_get_general_bounds()
    assert problem.provides_check()



from dataclasses import dataclass
import alpaqa as pa
import casadi as cs
import numpy as np
import numpy.random as nprand
import numpy.linalg as la
import alpaqa.casadi_loader as cl
import scipy.sparse as sp
import os

@dataclass
class OCProblemData:
    f: cs.Function
    h: cs.Function
    h_N: cs.Function
    l: cs.Function
    l_N: cs.Function
    c: cs.Function
    N: int

def convert_to_single_shooting(ocp: OCProblemData):
    nx = ocp.f.size1_in(0)
    nu = ocp.f.size1_in(1)
    mpc_x0 = cs.SX.sym("x0", nx)
    mpc_u = cs.SX.sym("u", (nu, ocp.N))  # Inputs
    mpc_x = ocp.f.mapaccum(ocp.N)(mpc_x0, mpc_u)  # Simulated states

    mpc_h = ocp.h.map(ocp.N)(cs.horzcat(mpc_x0, mpc_x[:, :-1]), mpc_u)
    mpc_stage_costs = cs.sum2(ocp.l.map(ocp.N)(mpc_h))
    mpc_terminal_cost = ocp.l_N(ocp.h_N(mpc_x[:, -1]))
    mpc_cost = cs.Function('f_mpc', [cs.vec(mpc_u), mpc_x0],
        [mpc_stage_costs + mpc_terminal_cost])

    g = [ocp.c(mpc_x0)]
    g += [ocp.c(mpc_x[:, i]) for i in range(ocp.N - 1)]
    # TODO: separate terminal constraint
    g += [ocp.c(mpc_x[:, ocp.N - 1])]
    mpc_constr = cs.Function("g", [cs.vec(mpc_u), mpc_x0],
        [cs.vertcat(*g)])

    return mpc_cost, mpc_constr

def convert_to_multiple_shooting_alm(ocp: OCProblemData, y, μ, D, D_N):
    N = ocp.N
    nx = ocp.f.size1_in(0)
    nu = ocp.f.size1_in(1)
    mpc_u = cs.SX.sym("u", (nu, N))  # Inputs
    mpc_x = cs.SX.sym("x", (nx, N + 1)) # States

    mpc_h = ocp.h.map(N)(mpc_x[:, :-1], mpc_u)
    mpc_stage_costs = cs.sum2(ocp.l.map(N)(mpc_h))
    mpc_terminal_cost = ocp.l_N(ocp.h_N(mpc_x[:, -1]))
    mpc_xu = cs.vertcat(cs.vec(cs.vertcat(mpc_x[:, :-1], mpc_u)), mpc_x[:, -1])

    penalty = 0
    for i in range(N):
        proj_D = lambda g: cs.fmax(D.lower, cs.fmin(g, D.upper))
        ζ = ocp.c(mpc_x[:, i]) + y[i * nx:i * nx + nx] / μ[i]
        penalty += 0.5 * μ[i] * cs.sum1((ζ - proj_D(ζ))**2)
    proj_D_N = lambda g: cs.fmax(D_N.lower, cs.fmin(g, D_N.upper))
    ζ = ocp.c(mpc_x[:, N]) + y[N * nx:N * nx + nx] / μ[N]
    penalty += 0.5 * μ[N] * cs.sum1((ζ - proj_D_N(ζ))**2)

    mpc_cost = cs.Function('f_mpc', [mpc_xu],
        [mpc_stage_costs + mpc_terminal_cost + penalty])

    return mpc_cost

def test_lqr():
    if not pa.with_casadi_ocp:
        return
    tol = 1e-10
    rng = nprand.default_rng(112233)

    # States and inputs
    N = 51
    nx, nu = 5, 7
    nc = nc_N = nx
    x_sym = cs.SX.sym("x", nx)
    u_sym = cs.SX.sym("u", nu)

    # Dynamics
    x_next = cs.vertcat(
        cs.tanh(x_sym[0]) + u_sym[0] * u_sym[1],
        cs.tanh(x_sym[1]) * cs.tanh(x_sym[0]) + u_sym[1] * u_sym[2],
        cs.tanh(x_sym[2]) * cs.tanh(x_sym[1]) + u_sym[3] * u_sym[4],
        cs.tanh(x_sym[3]) * cs.tanh(x_sym[2]) + u_sym[1] * u_sym[5],
        cs.tanh(x_sym[4]) * cs.tanh(x_sym[3]) + u_sym[2] * u_sym[6],
    )
    f = cs.Function("f", [x_sym, u_sym],
        [x_next])
    # Output
    h = cs.Function("h", [x_sym, u_sym],
        [cs.vertcat(cs.sinh(x_sym), 1. / (u_sym**2 + 1.))])
    hN = cs.Function("hN", [x_sym],
        [1. / (x_sym**2 + 1.)])
    # Cost
    L = rng.uniform(-1, 1, (nx + nu, nx + nu))
    L = L.T @ L
    hsym = cs.SX.sym("h", nx + nu)
    l = cs.Function("l", [hsym], 
        [cs.log(cs.cosh(hsym).T @ L @ cs.cosh(hsym))])
    LN = rng.uniform(-1, 1, (nx, nx))
    LN = LN.T @ LN
    hNsym = cs.SX.sym("hN", nx)
    lN = cs.Function("lN", [hNsym],
        [cs.log(cs.cosh(hNsym).T @ LN @ cs.cosh(hNsym))])
    # Constraints
    c = cs.Function("c", [x_sym],
        [cs.sinh(x_sym)])

    # Compile problem
    problem = cl.generate_and_compile_casadi_control_problem(
        N, f=f, l=l, l_N = lN, h=h, h_N=hN, c=c, c_N=c)
    
    # Convert to single-shooting problem
    ocp = OCProblemData(f=f, l=l, l_N=lN, h=h, h_N=hN, c=c, N=N)
    ss_cost, ss_constr = convert_to_single_shooting(ocp)
    cs_u = ss_cost.sx_in(0)

    # Random initialization
    u = rng.uniform(-2, 2, (N * nu,))
    y = rng.uniform(-0.5, 0.5, (nx * N + nx,))
    μ = rng.exponential(1, (nx * N + nx,))
    problem.x_init = rng.uniform(-1, 1, (nx,))

    # Check cost and gradient
    eval = pa.OCPEvaluator(problem)
    V, grad = eval.forward_backward(u, y, μ)

    cs_V = ss_cost(u, problem.x_init)
    assert abs(V - cs_V) < tol

    cs_grad = cs.Function("gr",
        [cs_u],
        [cs.gradient(ss_cost(cs_u, problem.x_init), cs_u)])(u)
    assert la.norm(grad - cs_grad) < tol

    # Check cost and gradient with constraints
    problem.U.lower = rng.uniform(-2.5, 0, nu)
    problem.U.upper = rng.uniform(0, +2.5, nu)
    problem.D.lower = rng.uniform(-1, -0.5, nx)
    problem.D.upper = rng.uniform(+0.5, +1, nx)
    problem.D_N.lower = problem.D.lower[::-1]
    problem.D_N.upper = problem.D.upper[::-1]

    aug_eval = pa.OCPEvaluator(problem)
    aug_V, aug_grad = aug_eval.forward_backward(u, y, μ)

    ss_D_lb = np.concatenate((np.tile(problem.D.lower, N), problem.D_N.lower))
    ss_D_ub = np.concatenate((np.tile(problem.D.upper, N), problem.D_N.upper))
    ss_μ = μ.copy()
    proj_D = lambda g: cs.fmax(ss_D_lb, cs.fmin(g, ss_D_ub))
    ζ = ss_constr(cs_u, problem.x_init) + y / ss_μ
    cs_aug_lagr = cs.Function("al", [cs_u],
        [ss_cost(cs_u, problem.x_init) + 
         0.5 * cs.sum1(ss_μ * (ζ - proj_D(ζ))**2)])

    cs_aug_V = cs_aug_lagr(u)
    assert abs(aug_V - cs_aug_V) < tol

    cs_aug_grad = cs.Function("algr", [cs_u],
        [cs.gradient(cs_aug_lagr(cs_u), cs_u)])(u).full().ravel()
    print("‖∇ψ(C++) - ∇ψ(CasADi) =", la.norm(aug_grad - cs_aug_grad))
    assert la.norm(aug_grad - cs_aug_grad) < tol

    # Simulate
    nxu = nx + nu
    xu = np.nan * np.ones((nxu * N + nx,))
    xu[0:nx] = problem.x_init
    for i in range(N):
        xi = xu[i * nxu:i * nxu + nx]
        ui = u[i * nu:i * nu + nu]
        xu[i * nxu + nx:i * nxu + nxu] = ui
        xu[(i + 1) * nxu:(i + 1) * nxu + nx] = f(xi, ui).full().ravel()

    # Build QP
    n_qp = nxu * N + nx
    Q_qp = np.zeros((n_qp, n_qp))
    for i in range(N):
        Qk = aug_eval.Qk(i, u, y, μ)
        Rk = aug_eval.Rk(i, u, np.arange(nu))
        Sk = aug_eval.Sk(i, u, np.arange(nu))
        Q_qp[i * nxu:i * nxu + nx, i * nxu:i * nxu + nx] = Qk
        Q_qp[i * nxu + nx:i * nxu + nxu, i * nxu + nx:i * nxu + nxu] = Rk
        Q_qp[i * nxu + nx:i * nxu + nxu, i * nxu:i * nxu + nx] = Sk
        Q_qp[i * nxu:i * nxu + nx, i * nxu + nx:i * nxu + nxu] = Sk.T
    Q_qp[N * nxu:N * nxu + nx, N * nxu:N * nxu + nx] = aug_eval.Qk(N, u, y, μ)

    ms_cost_alm = convert_to_multiple_shooting_alm(ocp, y, μ, problem.D, problem.D_N)
    cs_xu = ms_cost_alm.sx_in(0)
    # Compute true Hessian
    cs_hess_qp, _ = cs.hessian(ms_cost_alm(cs_xu), cs_xu)
    cs_hess_qp = cs.evalf(cs.substitute(cs_hess_qp, cs_xu, xu)).full()

    # Gauss-Newton Hessian approximation
    cs_Q_qp = np.zeros((n_qp, n_qp))
    cs_qr_qp = np.nan * np.ones((N * nxu + nx,))
    for i in range(N):
        xi = xu[i * nxu:i * nxu + nx]
        ui = xu[i * nxu + nx:i * nxu + nxu]
        μi = μ[i * nc:i * nc + nc]
        Jhx = cs.substitute(cs.jacobian(h(x_sym, ui), x_sym), x_sym, xi)
        Jhu = cs.substitute(cs.jacobian(h(xi, u_sym), u_sym), u_sym, ui)
        Lhh = cs.substitute(cs.hessian(l(hsym), hsym)[0], hsym, h(xi, ui))
        proj_D = lambda g: cs.fmax(problem.D.lower, cs.fmin(g, problem.D.upper))
        c_sym = cs.SX.sym("c", nx)
        ζ = c_sym + y[i * nx:i * nx + nx] / μi
        penalty = 0.5 * cs.sum1(μi * (ζ - proj_D(ζ))**2)
        ci = ocp.c(xi)
        M = cs.substitute(cs.hessian(penalty, c_sym)[0], c_sym, ci)
        Jc = cs.substitute(cs.jacobian(ocp.c(x_sym), x_sym), x_sym, xi)
        Qk = Jhx.T @ Lhh @ Jhx + Jc.T @ M @ Jc
        Rk = Jhu.T @ Lhh @ Jhu
        Sk = Jhu.T @ Lhh @ Jhx
        cs_Q_qp[i * nxu:i * nxu + nx, i * nxu:i * nxu + nx] = cs.evalf(Qk)
        cs_Q_qp[i * nxu + nx:i * nxu + nxu, i * nxu + nx:i * nxu + nxu] = cs.evalf(Rk)
        cs_Q_qp[i * nxu + nx:i * nxu + nxu, i * nxu:i * nxu + nx] = cs.evalf(Sk)
        cs_Q_qp[i * nxu:i * nxu + nx, i * nxu + nx:i * nxu + nxu] = cs.evalf(Sk).T
        grad_l = cs.substitute(cs.gradient(l(hsym), hsym), hsym, h(xi, ui))
        ζ = ci + y[i * nx:i * nx + nx] / μi
        qi = Jhx.T @ grad_l + Jc.T @ (μi * (ζ - proj_D(ζ)))
        ri = Jhu.T @ grad_l
        cs_qr_qp[i * nxu:i * nxu + nx] = cs.evalf(qi).full().ravel()
        cs_qr_qp[i * nxu + nx:i * nxu + nxu] = cs.evalf(ri).full().ravel()
    μN = μ[N * nc:N * nc + nc_N]
    xN = xu[N * nxu:N * nxu + nx]
    JhN = cs.substitute(cs.jacobian(hN(x_sym), x_sym), x_sym, xN)
    LNhh = cs.substitute(cs.hessian(lN(hNsym), hNsym)[0], hNsym, hN(xN))
    proj_D_N = lambda g: cs.fmax(problem.D_N.lower, cs.fmin(g, problem.D_N.upper))
    c_sym = cs.SX.sym("c", nx) # TODO: terminal constraints
    ζ = c_sym + y[N * nx:N * nx + nx] / μN
    penalty = 0.5 * cs.sum1(μN * (ζ - proj_D_N(ζ))**2)
    cN = ocp.c(xN)
    MN = cs.substitute(cs.hessian(penalty, c_sym)[0], c_sym, cN)
    JcN = cs.substitute(cs.jacobian(ocp.c(x_sym), x_sym), x_sym, xN)
    QN = JhN.T @ LNhh @ JhN + JcN.T @ MN @ JcN
    cs_Q_qp[N * nxu:N * nxu + nx, N * nxu:N * nxu + nx] = cs.evalf(QN)
    grad_lN = cs.substitute(cs.gradient(lN(hNsym), hNsym), hNsym, hN(xN))
    ζ = cN + y[N * nx:N * nx + nx] / μN
    qN = JhN.T @ grad_lN + JcN.T @ (μN * (ζ - proj_D_N(ζ)))
    cs_qr_qp[N * nxu:N * nxu + nx] = cs.evalf(qN).full().ravel()

    print("‖Q(C++) - Q(CasADi)‖ =", la.norm(Q_qp - cs_Q_qp))
    if "PYTEST_CURRENT_TEST" not in os.environ:
        import matplotlib.pyplot as plt
        plt.figure()
        plt.imshow(np.log10(abs(Q_qp - cs_Q_qp)))
        plt.colorbar()
        plt.title("C++ vs CasADi GN Hessian")
        plt.show()
    assert la.norm(Q_qp - cs_Q_qp) < tol

    if "PYTEST_CURRENT_TEST" not in os.environ:
        plt.figure()
        plt.imshow(np.log10(abs(cs_hess_qp - cs_Q_qp)))
        plt.colorbar()
        plt.title("GN Hessian vs True Hessian")
        plt.show()

    # Build constraint matrix
    triplets = []
    def add_identity(r, c, n, fac, triplets):
        for i in range(n):
            triplets += [(r + i, c + i, fac)]
    def add_block(r, c, M, triplets):
        for col in range(M.shape[1]):
            for row in range(M.shape[0]):
                val = M[row, col]
                if val:
                    triplets += [(r + row, c + col, val)]
    # Dynamics
    for i in range(N + 1):
        add_identity(i * nx, i * nxu, nx, 1., triplets)
        if i > 0:
            j = i - 1
            xj = xu[j * nxu:j * nxu + nx]
            uj = xu[j * nxu + nx:j * nxu + nxu]
            Aj = cs.substitute(cs.jacobian(f(x_sym, uj), x_sym), x_sym, xj)
            Bj = cs.substitute(cs.jacobian(f(xj, u_sym), u_sym), u_sym, uj)
            add_block(i * nx, j * nxu, cs.evalf(-Aj).full(), triplets)
            add_block(i * nx, j * nxu + nx, cs.evalf(-Bj).full(), triplets)
    # Box constraints
    γ = 0.56789
    b = []
    m = 0
    for i in range(N):
        for j in range(nu):
            ui = u[i * nu + j]
            lb_j = problem.U.lower[j]
            ub_j = problem.U.upper[j]
            gs = ui - γ * aug_grad[i * nu + j]
            if gs <= lb_j:
                triplets += [(N * nx + nx + m, i * nxu + nx + j, 1.)]
                b += [lb_j - ui]
                m += 1
            elif gs >= ub_j:
                triplets += [(N * nx + nx + m, i * nxu + nx + j, 1.)]
                b += [ub_j - ui]
                m += 1
    triplets = np.array(triplets)
    A_qp = sp.csc_matrix((triplets[:, 2], (triplets[:, 0], triplets[:, 1])), shape=(N * nx + nx + m, N * nxu + nx))

    if "PYTEST_CURRENT_TEST" not in os.environ:
        plt.figure()
        plt.imshow(np.log10(abs(A_qp.toarray())))
        plt.colorbar()
        plt.title("Constraint matrix A")
        plt.show()

    # Solve the QP
    import qpalm
    qp_data = qpalm.Data(Q_qp.shape[0], A_qp.shape[0])
    qp_data.Q, qp_data.A, qp_data.q, qp_data.c = Q_qp, A_qp, cs_qr_qp, 0
    qp_data.bmin = np.concatenate((np.zeros((N * nx + nx,)), np.array(b)))
    qp_data.bmax = np.concatenate((np.zeros((N * nx + nx,)), np.array(b)))

    qp_settings = qpalm.Settings()
    qp_settings.eps_abs = tol * 1e-2
    qp_settings.eps_abs_in = tol * 1e-2
    qp_settings.eps_rel = tol * 1e-2
    qp_settings.eps_rel_in = tol * 1e-2
    qp_settings.eps_prim_inf = tol * 1e-2
    qp_settings.eps_dual_inf = tol * 1e-2
    qp_settings.enable_dual_termination = True
    qp_sol = qpalm.Solver(qp_data, qp_settings)
    qp_sol.solve()
    qp_delxu = qp_sol.solution.x
    assert la.norm(A_qp @ qp_delxu - qp_data.bmin) < tol
    qp_delu = np.reshape(qp_delxu[:-nx], (nxu, N), order='F')[nx:, :]
    assert la.norm(qp_delxu[0:nx]) < tol
    A0 = cs.evalf(cs.substitute(cs.jacobian(f(x_sym, u[:nu]), x_sym), x_sym, xu[:nx])).full()
    B0 = cs.evalf(cs.substitute(cs.jacobian(f(xu[:nx], u_sym), u_sym), u_sym, u[:nu])).full()
    assert la.norm(A0 @ qp_delxu[0:nx] + B0 @ qp_delxu[nx:nx+nu] - qp_delxu[nxu:nxu+nx]) < tol

    # Compare with the LQR solution in C++
    delu = np.reshape(aug_eval.lqr_factor_solve(u, γ, y, μ), (nu, N), order='F')

    # Manual LQR solution (does not handle constraints, unused)
    qN = cs_qr_qp[N * nxu:N * nxu + nx]
    QN = cs_Q_qp[N * nxu:N * nxu + nx, N * nxu:N * nxu + nx]
    s = qN
    P = QN
    K = [None] * N
    e = [None] * N
    A = [None] * N
    B = [None] * N
    for i in range(N - 1, -1, -1):
        xi = xu[i * nxu:i * nxu + nx]
        ui = xu[i * nxu + nx:i * nxu + nxu]
        qi = cs_qr_qp[i * nxu:i * nxu + nx]
        ri = cs_qr_qp[i * nxu + nx:i * nxu + nxu]
        Qi = cs_Q_qp[i * nxu:i * nxu + nx, i * nxu:i * nxu + nx]
        Ri = cs_Q_qp[i * nxu + nx:i * nxu + nxu, i * nxu + nx:i * nxu + nxu]
        Si = cs_Q_qp[i * nxu + nx:i * nxu + nxu, i * nxu:i * nxu + nx]
        Ai = cs.evalf(cs.substitute(cs.jacobian(f(x_sym, ui), x_sym), x_sym, xi)).full()
        Bi = cs.evalf(cs.substitute(cs.jacobian(f(xi, u_sym), u_sym), u_sym, ui)).full()
        R̅ = Ri + Bi.T @ P @ Bi
        S̅ = Si + Bi.T @ P @ Ai
        y = s
        t = Bi.T @ y + ri
        K[i] = -la.solve(R̅, S̅)
        e[i] = -la.solve(R̅, t)
        # print("R̅", R̅, "S̅", S̅, "y", y, "t", t, "Ki", K[i], "ei", e[i], sep='\n')
        s = S̅.T @ e[i] + Ai.T @ y + qi
        P = Qi + Ai.T @ P @ Ai + S̅.T @ K[i]
        A[i] = Ai
        B[i] = Bi
    py_delxu = np.zeros((N * nxu + nx,))
    for i in range(N):
        delxi = py_delxu[i * nxu:i * nxu + nx]
        delui = py_delxu[i * nxu + nx:i * nxu + nxu]
        delui[:] = K[i] @ delxi + e[i]
        py_delxu[(i+1) * nxu:(i+1) * nxu + nx] = A[i] @ delxi + B[i] @ delui

    # Show difference between QP and LQR solutions
    qp_delU = qp_delu.reshape(-1, order='F')
    delU = delu.reshape(-1, order='F')
    print("‖Δu(C++) - Δu(QPALM)‖ =", la.norm(delU - qp_delU))
    if "PYTEST_CURRENT_TEST" not in os.environ:
        plt.figure()
        plt.plot(qp_delU, '.-', label="QPALM")
        plt.plot(delU, '.-', label="C++ LQR")
        plt.title("QPALM solution vs C++ LQR solution")
        plt.legend()
        plt.show()
    assert la.norm(delU - qp_delU) < 1e3 * tol

if __name__ == '__main__':
    test_lqr()



import alpaqa as pa
import numpy as np
import numpy.linalg as la


def test_nuclear_norm():
    θ, φ = 0.3456, 0.8765
    U = np.array([[np.cos(θ), -np.sin(θ)], [np.sin(θ), np.cos(θ)]])
    V = np.array([[np.cos(φ), -np.sin(φ)], [np.sin(φ), np.cos(φ)]])
    Σ = np.diag([1, 0.05])
    x = U @ Σ @ V.T
    Σ_expected = np.diag([0.875, 0])
    y = np.empty((2, 2), order="F")
    h = pa.functions.NuclearNorm(0.25)
    hy = pa.prox(h, x, y, 0.5)
    assert abs(hy - 0.25 * 0.875) < 1e-12
    assert np.allclose(y, U @ Σ_expected @ V.T, rtol=1e-12, atol=1e-12)
    hy, y = pa.prox(h, x, 0.5)
    assert abs(hy - 0.25 * 0.875) < 1e-12
    assert np.allclose(y, U @ Σ_expected @ V.T, rtol=1e-12, atol=1e-12)


def test_nuclear_norm_reshape():
    θ, φ = 0.3456, 0.8765
    U = np.array([[np.cos(θ), -np.sin(θ)], [np.sin(θ), np.cos(θ)]])
    V = np.array([[np.cos(φ), -np.sin(φ)], [np.sin(φ), np.cos(φ)]])
    Σ = np.diag([1, 0.05])
    x = U @ Σ @ V.T
    Σ_expected = np.diag([0.875, 0])
    y = np.empty(2 * 2, order="F")
    h = pa.functions.NuclearNorm(0.25, 2, 2)
    hy = pa.prox(h, x.ravel("F"), y, 0.5)
    y = np.reshape(y, (2, 2), "F")
    assert abs(hy - 0.25 * 0.875) < 1e-12
    assert np.allclose(y, U @ Σ_expected @ V.T, rtol=1e-12, atol=1e-12)
    hy, y = pa.prox(h, x, 0.5)
    y = np.reshape(y, (2, 2), "F")
    assert abs(hy - 0.25 * 0.875) < 1e-12
    assert np.allclose(y, U @ Σ_expected @ V.T, rtol=1e-12, atol=1e-12)


def test_l1_norm():
    x = [-1, -0.125, -0.1, 0, 0.05, 0.12, 0.13, 1, 100]
    y_expected = np.array([-0.875, 0, 0, 0, 0, 0, 0.005, 0.875, 99.875])
    y = np.empty(len(x))
    λ = 0.25
    h = pa.functions.L1Norm(λ)
    hy = pa.prox(h, x, y, 0.5)
    assert np.allclose(y, y_expected, rtol=1e-12, atol=1e-12)
    assert abs(0.25 * la.norm(y_expected, 1) - hy) < 1e-12
    hy, y = pa.prox(h, x, 0.5)
    assert np.allclose(y.ravel(), y_expected, rtol=1e-12, atol=1e-12)
    assert abs(0.25 * la.norm(y_expected, 1) - hy) < 1e-12


def test_l1_norm_vec():
    x = [-1, -0.125, -0.1, 0, 0.05, 0.12, 0.13, 1, 100]
    y_expected = np.array([-0.875, 0, 0, 0, 0, 0, 0.005, 0.875, 99.875])
    y = np.empty(len(x))
    λ = 0.25 * np.ones(len(x))
    h = pa.functions.L1NormElementwise(λ)
    hy = pa.prox(h, x, y, 0.5)
    assert np.allclose(y, y_expected, rtol=1e-12, atol=1e-12)
    assert abs(0.25 * la.norm(y_expected, 1) - hy) < 1e-12
    hy, y = pa.prox(h, x, 0.5)
    assert np.allclose(y.ravel(), y_expected, rtol=1e-12, atol=1e-12)
    assert abs(0.25 * la.norm(y_expected, 1) - hy) < 1e-12


def test_l1_norm_step():
    rng = np.random.default_rng(4321)
    x_fw = np.array([-1, -0.125, -0.1, 0, 0.05, 0.12, 0.13, 1, 100])
    γ_fwd = 0.1234
    x = rng.uniform(-1, 1, len(x_fw))
    g = (x - x_fw) / γ_fwd
    assert np.allclose(x_fw, x - γ_fwd * g)
    y_expected = np.array([-0.875, 0, 0, 0, 0, 0, 0.005, 0.875, 99.875])
    y = np.empty(len(x))
    p = np.empty(len(x))
    h = pa.functions.L1Norm(0.25)
    hy = pa.prox_step(h, x, g, y, p, 0.5, -γ_fwd)
    assert np.allclose(y, y_expected, rtol=1e-12, atol=1e-12)
    assert abs(0.25 * la.norm(y_expected, 1) - hy) < 1e-12
    hy, y, p = pa.prox_step(h, x, g, 0.5, -γ_fwd)
    assert np.allclose(y.ravel(), y_expected, rtol=1e-12, atol=1e-12)
    assert abs(0.25 * la.norm(y_expected, 1) - hy) < 1e-12


def test_box():
    rng = np.random.default_rng(1234)
    n = 128
    x = rng.uniform(-1, 1, n)
    lb = rng.uniform(-1.5, 1, n)
    ub = lb + rng.uniform(0, 1, n)
    y_expected = np.fmax(lb, np.fmin(x, ub))
    y = np.empty(len(x))
    h = pa.Box(lower=lb, upper=ub)
    hy = pa.prox(h, x, y, 100)
    assert np.allclose(y, y_expected, rtol=1e-12, atol=1e-12)
    assert hy == 0
    hy, y = pa.prox(h, x, 100)
    assert np.allclose(y.ravel(), y_expected, rtol=1e-12, atol=1e-12)
    assert hy == 0


def test_box_matrix():
    rng = np.random.default_rng(1234)
    n = 128
    x = rng.uniform(-1, 1, (n, n))
    lb = rng.uniform(-1.5, 1, n * n)
    ub = lb + rng.uniform(0, 1, n * n)
    y = np.empty(x.shape, order="F")
    x_vec = x.ravel(order="F")
    y_expected = np.fmax(lb, np.fmin(x_vec, ub)).reshape((n, n), order="F")
    h = pa.Box(lower=lb, upper=ub)
    hy = pa.prox(h, x, y, 100)
    assert np.allclose(y, y_expected, rtol=1e-12, atol=1e-12)
    assert hy == 0
    hy, y = pa.prox(h, x, 100)
    assert np.allclose(y, y_expected, rtol=1e-12, atol=1e-12)
    assert hy == 0


def test_box_step():
    rng = np.random.default_rng(1234)
    n = 128
    x = rng.uniform(-1, 1, n)
    g = rng.uniform(-1, 1, n)
    lb = rng.uniform(-1.5, 1, n)
    ub = lb + rng.uniform(0, 1, n)
    y_expected = np.fmax(lb, np.fmin(x - 0.1 * g, ub))
    y = np.empty(len(x))
    p = np.empty(len(x))
    h = pa.Box(lower=lb, upper=ub)
    hy = pa.prox_step(h, x, g, y, p, 100, -0.1)
    p_expected = y_expected - x
    assert np.allclose(y, y_expected, rtol=1e-12, atol=1e-12)
    assert np.allclose(p, p_expected, rtol=1e-12, atol=1e-12)
    assert hy == 0
    hy, y, p = pa.prox_step(h, x, g, 100, -0.1)
    assert np.allclose(y.ravel(), y_expected, rtol=1e-12, atol=1e-12)
    assert np.allclose(p.ravel(), p_expected, rtol=1e-12, atol=1e-12)
    assert hy == 0


if __name__ == "__main__":
    test_nuclear_norm()
    test_nuclear_norm_reshape()
    test_l1_norm()
    test_l1_norm_vec()
    test_l1_norm_step()
    test_box()
    test_box_matrix()
    test_box_step()



"""
Python interface to alpaqa's C++ implementation.
"""
from __future__ import annotations
import datetime
import typing
from . import float64
from . import longdouble
__all__: list[str] = ['ApproxKKT', 'ApproxKKT2', 'BasedOnCurvature', 'BasedOnExternalStepSize', 'Busy', 'Converged', 'DynamicLoadFlags', 'EvalCounter', 'FPRNorm', 'FPRNorm2', 'Interrupted', 'Ipopt', 'LBFGSBpp', 'LBFGSStepsize', 'Lower', 'MaxIter', 'MaxTime', 'NoProgress', 'NotFinite', 'OCPEvalCounter', 'PANOCStopCrit', 'ProjGradNorm', 'ProjGradNorm2', 'ProjGradUnitNorm', 'ProjGradUnitNorm2', 'SolverStatus', 'Symmetry', 'Unsymmetric', 'Upper', 'build_time', 'commit_hash', 'float64', 'longdouble', 'not_implemented_error', 'with_casadi', 'with_casadi_ocp', 'with_external_casadi', 'with_ipopt']
class DynamicLoadFlags:
    """
    C++ documentation: :cpp:class:`guanaqo::DynamicLoadFlags`
    """
    deepbind: bool
    global_: bool
    lazy: bool
    nodelete: bool
    @typing.overload
    def __init__(self, params: dict) -> None:
        ...
    @typing.overload
    def __init__(self, **kwargs) -> None:
        ...
    def to_dict(self) -> dict:
        ...
class EvalCounter:
    """
    C++ documentation: :cpp:class:`alpaqa::EvalCounter`
    
    """
    class EvalTimer:
        """
        C++ documentation: :cpp:class:`alpaqa::EvalCounter::EvalTimer`
        
        """
        augmented_lagrangian: datetime.timedelta
        augmented_lagrangian_and_gradient: datetime.timedelta
        augmented_lagrangian_gradient: datetime.timedelta
        augmented_lagrangian_hessian: datetime.timedelta
        augmented_lagrangian_hessian_product: datetime.timedelta
        constraints: datetime.timedelta
        constraints_gradient_product: datetime.timedelta
        constraints_jacobian: datetime.timedelta
        grad_gi: datetime.timedelta
        inactive_indices_res_lna: datetime.timedelta
        lagrangian_gradient: datetime.timedelta
        lagrangian_hessian: datetime.timedelta
        lagrangian_hessian_product: datetime.timedelta
        objective: datetime.timedelta
        objective_and_constraints: datetime.timedelta
        objective_and_gradient: datetime.timedelta
        objective_gradient: datetime.timedelta
        objective_gradient_and_constraints_gradient_product: datetime.timedelta
        projecting_difference_constraints: datetime.timedelta
        projection_multipliers: datetime.timedelta
        proximal_gradient_step: datetime.timedelta
        def __getstate__(self) -> tuple:
            ...
        def __setstate__(self, arg0: tuple) -> None:
            ...
    augmented_lagrangian: int
    augmented_lagrangian_and_gradient: int
    augmented_lagrangian_gradient: int
    augmented_lagrangian_hessian: int
    augmented_lagrangian_hessian_product: int
    constraints: int
    constraints_gradient_product: int
    constraints_jacobian: int
    grad_gi: int
    inactive_indices_res_lna: int
    lagrangian_gradient: int
    lagrangian_hessian: int
    lagrangian_hessian_product: int
    objective: int
    objective_and_constraints: int
    objective_and_gradient: int
    objective_gradient: int
    objective_gradient_and_constraints_gradient_product: int
    projecting_difference_constraints: int
    projection_multipliers: int
    proximal_gradient_step: int
    time: EvalCounter.EvalTimer
    def __getstate__(self) -> tuple:
        ...
    def __setstate__(self, arg0: tuple) -> None:
        ...
    def __str__(self) -> str:
        ...
class LBFGSStepsize:
    """
    C++ documentation: :cpp:enum:`alpaqa::LBFGSStepSize`
    
    Members:
    
      BasedOnExternalStepSize
    
      BasedOnCurvature
    """
    BasedOnCurvature: typing.ClassVar[LBFGSStepsize]  # value = <LBFGSStepsize.BasedOnCurvature: 1>
    BasedOnExternalStepSize: typing.ClassVar[LBFGSStepsize]  # value = <LBFGSStepsize.BasedOnExternalStepSize: 0>
    __members__: typing.ClassVar[dict[str, LBFGSStepsize]]  # value = {'BasedOnExternalStepSize': <LBFGSStepsize.BasedOnExternalStepSize: 0>, 'BasedOnCurvature': <LBFGSStepsize.BasedOnCurvature: 1>}
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: int) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: int) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class OCPEvalCounter:
    """
    C++ documentation: :cpp:class:`alpaqa::OCPEvalCounter`
    
    """
    class OCPEvalTimer:
        """
        C++ documentation: :cpp:class:`alpaqa::OCPEvalCounter::OCPEvalTimer`
        
        """
        add_Q: datetime.timedelta
        add_Q_N: datetime.timedelta
        add_R_masked: datetime.timedelta
        add_R_prod_masked: datetime.timedelta
        add_S_masked: datetime.timedelta
        add_S_prod_masked: datetime.timedelta
        add_gn_hess_constr: datetime.timedelta
        add_gn_hess_constr_N: datetime.timedelta
        constr: datetime.timedelta
        constr_N: datetime.timedelta
        f: datetime.timedelta
        grad_constr_prod: datetime.timedelta
        grad_constr_prod_N: datetime.timedelta
        grad_f_prod: datetime.timedelta
        h: datetime.timedelta
        h_N: datetime.timedelta
        jac_f: datetime.timedelta
        l: datetime.timedelta
        l_N: datetime.timedelta
        q_N: datetime.timedelta
        qr: datetime.timedelta
        def __getstate__(self) -> tuple:
            ...
        def __setstate__(self, arg0: tuple) -> None:
            ...
    add_Q: int
    add_Q_N: int
    add_R_masked: int
    add_R_prod_masked: int
    add_S_masked: int
    add_S_prod_masked: int
    add_gn_hess_constr: int
    add_gn_hess_constr_N: int
    constr: int
    constr_N: int
    f: int
    grad_constr_prod: int
    grad_constr_prod_N: int
    grad_f_prod: int
    h: int
    h_N: int
    jac_f: int
    l: int
    l_N: int
    q_N: int
    qr: int
    time: OCPEvalCounter.OCPEvalTimer
    def __getstate__(self) -> tuple:
        ...
    def __setstate__(self, arg0: tuple) -> None:
        ...
    def __str__(self) -> str:
        ...
class PANOCStopCrit:
    """
    C++ documentation: :cpp:enum:`alpaqa::PANOCStopCrit`
    
    Members:
    
      ApproxKKT
    
      ApproxKKT2
    
      ProjGradNorm
    
      ProjGradNorm2
    
      ProjGradUnitNorm
    
      ProjGradUnitNorm2
    
      FPRNorm
    
      FPRNorm2
    
      Ipopt
    
      LBFGSBpp
    """
    ApproxKKT: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.ApproxKKT: 0>
    ApproxKKT2: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.ApproxKKT2: 1>
    FPRNorm: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.FPRNorm: 6>
    FPRNorm2: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.FPRNorm2: 7>
    Ipopt: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.Ipopt: 8>
    LBFGSBpp: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.LBFGSBpp: 9>
    ProjGradNorm: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.ProjGradNorm: 2>
    ProjGradNorm2: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.ProjGradNorm2: 3>
    ProjGradUnitNorm: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.ProjGradUnitNorm: 4>
    ProjGradUnitNorm2: typing.ClassVar[PANOCStopCrit]  # value = <PANOCStopCrit.ProjGradUnitNorm2: 5>
    __members__: typing.ClassVar[dict[str, PANOCStopCrit]]  # value = {'ApproxKKT': <PANOCStopCrit.ApproxKKT: 0>, 'ApproxKKT2': <PANOCStopCrit.ApproxKKT2: 1>, 'ProjGradNorm': <PANOCStopCrit.ProjGradNorm: 2>, 'ProjGradNorm2': <PANOCStopCrit.ProjGradNorm2: 3>, 'ProjGradUnitNorm': <PANOCStopCrit.ProjGradUnitNorm: 4>, 'ProjGradUnitNorm2': <PANOCStopCrit.ProjGradUnitNorm2: 5>, 'FPRNorm': <PANOCStopCrit.FPRNorm: 6>, 'FPRNorm2': <PANOCStopCrit.FPRNorm2: 7>, 'Ipopt': <PANOCStopCrit.Ipopt: 8>, 'LBFGSBpp': <PANOCStopCrit.LBFGSBpp: 9>}
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: int) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: int) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class SolverStatus:
    """
    C++ documentation: :cpp:enum:`alpaqa::SolverStatus`
    
    Members:
    
      Busy : In progress.
    
      Converged : Converged and reached given tolerance
    
      MaxTime : Maximum allowed execution time exceeded
    
      MaxIter : Maximum number of iterations exceeded
    
      NotFinite : Intermediate results were infinite or NaN
    
      NoProgress : No progress was made in the last iteration
    
      Interrupted : Solver was interrupted by the user
    """
    Busy: typing.ClassVar[SolverStatus]  # value = <SolverStatus.Busy: 0>
    Converged: typing.ClassVar[SolverStatus]  # value = <SolverStatus.Converged: 1>
    Interrupted: typing.ClassVar[SolverStatus]  # value = <SolverStatus.Interrupted: 6>
    MaxIter: typing.ClassVar[SolverStatus]  # value = <SolverStatus.MaxIter: 3>
    MaxTime: typing.ClassVar[SolverStatus]  # value = <SolverStatus.MaxTime: 2>
    NoProgress: typing.ClassVar[SolverStatus]  # value = <SolverStatus.NoProgress: 5>
    NotFinite: typing.ClassVar[SolverStatus]  # value = <SolverStatus.NotFinite: 4>
    __members__: typing.ClassVar[dict[str, SolverStatus]]  # value = {'Busy': <SolverStatus.Busy: 0>, 'Converged': <SolverStatus.Converged: 1>, 'MaxTime': <SolverStatus.MaxTime: 2>, 'MaxIter': <SolverStatus.MaxIter: 3>, 'NotFinite': <SolverStatus.NotFinite: 4>, 'NoProgress': <SolverStatus.NoProgress: 5>, 'Interrupted': <SolverStatus.Interrupted: 6>}
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __ge__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __gt__(self, other: typing.Any) -> bool:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: int) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __le__(self, other: typing.Any) -> bool:
        ...
    def __lt__(self, other: typing.Any) -> bool:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: int) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class Symmetry:
    """
    C++ documentation: :cpp:enum:`alpaqa::sparsity::Symmetry`
    
    Members:
    
      Unsymmetric
    
      Upper
    
      Lower
    """
    Lower: typing.ClassVar[Symmetry]  # value = <Symmetry.Lower: 2>
    Unsymmetric: typing.ClassVar[Symmetry]  # value = <Symmetry.Unsymmetric: 0>
    Upper: typing.ClassVar[Symmetry]  # value = <Symmetry.Upper: 1>
    __members__: typing.ClassVar[dict[str, Symmetry]]  # value = {'Unsymmetric': <Symmetry.Unsymmetric: 0>, 'Upper': <Symmetry.Upper: 1>, 'Lower': <Symmetry.Lower: 2>}
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: int) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: int) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class not_implemented_error(NotImplementedError):
    pass
ApproxKKT: PANOCStopCrit  # value = <PANOCStopCrit.ApproxKKT: 0>
ApproxKKT2: PANOCStopCrit  # value = <PANOCStopCrit.ApproxKKT2: 1>
BasedOnCurvature: LBFGSStepsize  # value = <LBFGSStepsize.BasedOnCurvature: 1>
BasedOnExternalStepSize: LBFGSStepsize  # value = <LBFGSStepsize.BasedOnExternalStepSize: 0>
Busy: SolverStatus  # value = <SolverStatus.Busy: 0>
Converged: SolverStatus  # value = <SolverStatus.Converged: 1>
FPRNorm: PANOCStopCrit  # value = <PANOCStopCrit.FPRNorm: 6>
FPRNorm2: PANOCStopCrit  # value = <PANOCStopCrit.FPRNorm2: 7>
Interrupted: SolverStatus  # value = <SolverStatus.Interrupted: 6>
Ipopt: PANOCStopCrit  # value = <PANOCStopCrit.Ipopt: 8>
LBFGSBpp: PANOCStopCrit  # value = <PANOCStopCrit.LBFGSBpp: 9>
Lower: Symmetry  # value = <Symmetry.Lower: 2>
MaxIter: SolverStatus  # value = <SolverStatus.MaxIter: 3>
MaxTime: SolverStatus  # value = <SolverStatus.MaxTime: 2>
NoProgress: SolverStatus  # value = <SolverStatus.NoProgress: 5>
NotFinite: SolverStatus  # value = <SolverStatus.NotFinite: 4>
ProjGradNorm: PANOCStopCrit  # value = <PANOCStopCrit.ProjGradNorm: 2>
ProjGradNorm2: PANOCStopCrit  # value = <PANOCStopCrit.ProjGradNorm2: 3>
ProjGradUnitNorm: PANOCStopCrit  # value = <PANOCStopCrit.ProjGradUnitNorm: 4>
ProjGradUnitNorm2: PANOCStopCrit  # value = <PANOCStopCrit.ProjGradUnitNorm2: 5>
Unsymmetric: Symmetry  # value = <Symmetry.Unsymmetric: 0>
Upper: Symmetry  # value = <Symmetry.Upper: 1>
__version__: str = '1.1.0a1'
build_time: str = '2025-09-03T21:12:52Z'
commit_hash: str = '2cc06d006231f4ec254f89b61a3ee14d3621af4d'
with_casadi: bool = True
with_casadi_ocp: bool = False
with_external_casadi: bool = True
with_ipopt: bool = True



from copy import copy, deepcopy
import pickle
import alpaqa as pa
import numpy as np
from pprint import pprint
import pytest


@pytest.mark.skipif(not pa.with_casadi, reason="requires CasADi")
def test_alm():
    import alpaqa.casadi_loader as cl

    pp = pa.PANOCParams()
    lbfgs = pa.LBFGSDirection()
    panoc = pa.PANOCSolver(pp, lbfgs)
    pprint(panoc.direction.params)
    solver = pa.ALMSolver(pa.ALMParams(), panoc)
    print(f"Solver: {solver} "
          f"({solver.__class__.__module__}.{solver.__class__.__qualname__})")

    import casadi as cs

    n = 2
    m = 2
    x = cs.SX.sym("x", n)

    Q = np.array([[1.5, 0.5], [0.5, 1.5]])
    f_ = 0.5 * x.T @ Q @ x
    g_ = x
    f = cs.Function("f", [x], [f_])
    g = cs.Function("g", [x], [g_])

    name = "testproblem"
    p = cl.generate_and_compile_casadi_problem(f, g, name=name)
    p = copy(p) # test copying/cloning
    p = deepcopy(p) # test copying/cloning
    print(p)
    print("C", p.variable_bounds.lower, p.variable_bounds.upper)
    print("D", p.general_bounds.lower, p.general_bounds.upper)
    p.general_bounds.lower = [-np.inf, 0.5]
    p.general_bounds.upper = [+np.inf, +np.inf]
    solver = pa.PANOCSolver(
        pa.PANOCParams(max_iter=200, print_interval=1),
        pa.LBFGSDirection(pa.LBFGS.Params(memory=5)),
    )
    almparams = pa.ALMParams(max_iter=20, print_interval=1, tolerance=1e-12, dual_tolerance=1e-12)
    almsolver = pa.ALMSolver(almparams, solver)
    cnt = pa.problem_with_counters(p)
    x0 = np.array([3, 3])
    y0 = np.zeros((m, ))
    
    x, y, stats = almsolver(cnt.problem, x=x0, y=y0)

    print()
    print(cnt.evaluations)
    print(stats["status"])
    print("x", x)
    print("y", y)
    pprint(stats)
    assert stats['status'] == pa.SolverStatus.Converged
    assert np.linalg.norm(x - [-1/6, 0.5]) < 1e-5
    assert np.linalg.norm(y - [0, -2/3]) < 1e-5


@pytest.mark.skipif(not pa.with_casadi, reason="requires CasADi")
def test_alm_pyapi_compile():
    import casadi as cs

    n = 2
    m = 2
    x = cs.SX.sym("x", n)

    Q = np.array([[1.5, 0.5], [0.5, 1.5]])
    f = 0.5 * x.T @ Q @ x
    g = x
    D = [-np.inf, 0.5], [+np.inf, +np.inf]
    p = pa.minimize(f, x).subject_to(g, D).compile()
    p = copy(p) # test copying/cloning
    p = deepcopy(p) # test copying/cloning
    print(p)
    solver = pa.PANOCSolver(
        pa.PANOCParams(max_iter=200, print_interval=1),
        pa.LBFGSDirection(pa.LBFGS.Params(memory=5)),
    )
    almparams = pa.ALMParams(max_iter=20, print_interval=1, tolerance=1e-12, dual_tolerance=1e-12)
    almsolver = pa.ALMSolver(almparams, solver)
    cnt = pa.problem_with_counters(p)
    x0 = np.array([3, 3])
    y0 = np.zeros((m, ))
    
    x, y, stats = almsolver(cnt.problem, x=x0, y=y0)

    print()
    print(cnt.evaluations)
    print(stats["status"])
    print("x", x)
    print("y", y)
    pprint(stats)
    assert stats['status'] == pa.SolverStatus.Converged
    assert np.linalg.norm(x - [-1/6, 0.5]) < 1e-5
    assert np.linalg.norm(y - [0, -2/3]) < 1e-5


@pytest.mark.skipif(not pa.with_external_casadi, reason="requires CasADi")
def test_alm_pyapi_build():
    import casadi as cs

    n = 2
    m = 2
    x = cs.SX.sym("x", n)

    Q = np.array([[1.5, 0.5], [0.5, 1.5]])
    f = 0.5 * x.T @ Q @ x
    g = x
    D = [-np.inf, 0.5], [+np.inf, +np.inf]
    p = pa.minimize(f, x).subject_to(g, D).build()
    p = copy(p) # test copying/cloning
    p = deepcopy(p) # test copying/cloning
    print(p)
    solver = pa.PANOCSolver(
        pa.PANOCParams(max_iter=200, print_interval=1),
        pa.LBFGSDirection(pa.LBFGS.Params(memory=5)),
    )
    almparams = pa.ALMParams(max_iter=20, print_interval=1, tolerance=1e-12, dual_tolerance=1e-12)
    almsolver = pa.ALMSolver(almparams, solver)
    cnt = pa.problem_with_counters(p)
    x0 = np.array([3, 3])
    y0 = np.zeros((m, ))
    
    x, y, stats = almsolver(cnt.problem, x=x0, y=y0)

    print()
    print(cnt.evaluations)
    print(stats["status"])
    print("x", x)
    print("y", y)
    pprint(stats)
    assert stats['status'] == pa.SolverStatus.Converged
    assert np.linalg.norm(x - [-1/6, 0.5]) < 1e-5
    assert np.linalg.norm(y - [0, -2/3]) < 1e-5


class MyProblem(pa.BoxConstrProblem):
    vec = np.ndarray
    Q = np.array([[1.5, 0.5], [0.5, 1.5]])
    A = np.eye(2)

    def __init__(self) -> None:
        super().__init__(self.Q.shape[0], self.A.shape[0])

    def eval_objective(self, x: vec) -> float:
        return 0.5 * x.T @ self.Q @ x

    def eval_objective_gradient(self, x: vec, grad_fx: vec):
        grad_fx[:] = self.Q @ x

    def eval_constraints(self, x: vec, gx: vec):
        gx[:] = self.A @ x

    def eval_constraints_gradient_product(self, x: vec, y: vec, g: vec):
        g[:] = self.A.T @ y


def get_pickled_problem():
    p = MyProblem()
    p.general_bounds.lower = [-np.inf, 0.5]
    p.general_bounds.upper = [+np.inf, +np.inf]
    return pickle.dumps(p)


def test_alm_inherit():

    p = pickle.loads(get_pickled_problem())
    solver = pa.PANOCSolver(
        pa.PANOCParams(max_iter=200, print_interval=1),
        pa.LBFGSDirection(pa.LBFGS.Params(memory=5)),
    )
    almparams = pa.ALMParams(max_iter=20, print_interval=1, tolerance=1e-12, dual_tolerance=1e-12)
    almsolver = pa.ALMSolver(almparams, solver)
    cnt = pa.problem_with_counters(p)
    x0 = np.array([3, 3])
    y0 = np.array([0, 0])
    x, y, stats = almsolver(pa.Problem(p), x=x0, y=y0)
    x, y, stats = almsolver(cnt.problem, x=x0, y=y0)

    print()
    print(cnt.evaluations)
    print(stats["status"])
    print("x", x)
    print("y", y)
    pprint(stats)
    assert stats['status'] == pa.SolverStatus.Converged
    assert np.linalg.norm(x - [-1/6, 0.5]) < 1e-5
    assert np.linalg.norm(y - [0, -2/3]) < 1e-5


def test_alm_structured_inherit():

    p = pickle.loads(get_pickled_problem())
    solver = pa.PANOCSolver(
        pa.PANOCParams(max_iter=200, print_interval=1),
        pa.StructuredLBFGSDirection(pa.LBFGS.Params(memory=5)),
    )
    pprint(solver.direction.params)
    almparams = pa.ALMParams(max_iter=20, print_interval=1, tolerance=1e-12, dual_tolerance=1e-12)
    almsolver = pa.ALMSolver(almparams, solver)
    cnt = pa.problem_with_counters(p)
    x0 = np.array([3, 3])
    y0 = np.array([0, 0])
    x, y, stats = almsolver(pa.Problem(p), x=x0, y=y0)
    x, y, stats = almsolver(cnt.problem, x=x0, y=y0)

    print()
    print(almsolver)
    print(cnt.evaluations)
    print(stats["status"])
    print("x", x)
    print("y", y)
    pprint(stats)
    assert stats['status'] == pa.SolverStatus.Converged
    assert np.linalg.norm(x - [-1/6, 0.5]) < 1e-5
    assert np.linalg.norm(y - [0, -2/3]) < 1e-5


if __name__ == '__main__':
    test_alm()
    test_alm_inherit()
    test_alm_structured_inherit()



"""
Augmented Lagrangian and PANOC solvers for nonconvex numerical optimization.
"""

__version__ = "1.1.0a1"

import contextlib
from .alpaqa import *
from .alpaqa import __c_version__

assert __version__ == __c_version__

with contextlib.suppress(ModuleNotFoundError):  # Don't fail if CasADi is unavailable
    from .pyapi import *

# For Sphinx
__all__ = [v for v in dir() if not v.startswith("_") and v != "alpaqa"] + [
    "__c_version__"
]



from dataclasses import dataclass
import casadi as cs
from typing import Union, Optional, Tuple
import numpy as np
from copy import copy
from ..casadi_loader import generate_and_compile_casadi_problem
from ..casadi_generator import _prepare_casadi_problem
from ..alpaqa import CasADiProblem, deserialize_casadi_problem
import inspect


@dataclass
class MinimizationProblemDescription:
    """
    High-level description of a minimization problem.
    """

    objective_expr: Union[cs.SX, cs.MX]
    variable: Union[cs.SX, cs.MX]
    constraints_expr: Optional[Union[cs.SX, cs.MX]] = None
    penalty_constraints_expr: Optional[Union[cs.SX, cs.MX]] = None
    parameter: Optional[Union[cs.SX, cs.MX]] = None
    parameter_value: Optional[np.ndarray] = None
    regularizer: Optional[Union[float, np.ndarray]] = None
    bounds: Optional[Tuple[np.ndarray, np.ndarray]] = None
    constraints_bounds: Optional[Tuple[np.ndarray, np.ndarray]] = None
    penalty_constraints_bounds: Optional[Tuple[np.ndarray, np.ndarray]] = None
    name: str = "alpaqa_casadi_problem"

    @staticmethod
    def _assert_not_set_before(value):
        if value is not None:
            caller_frame = inspect.getouterframes(inspect.currentframe())[1]
            raise ValueError(f"{caller_frame.function} cannot be called twice")

    def subject_to_box(
        self, C: Tuple[np.ndarray, np.ndarray]
    ) -> "MinimizationProblemDescription":
        """
        Add box constraints :math:`x \\in C` on the problem variables.
        """
        self._assert_not_set_before(self.bounds)
        ret = copy(self)
        ret.bounds = C
        return ret

    def subject_to(
        self,
        g: Union[cs.SX, cs.MX],
        D: Optional[Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]] = None,
    ) -> "MinimizationProblemDescription":
        """
        Add general constraints :math:`g(x) \\in D`, handled using an augmented
        Lagrangian method.
        """
        self._assert_not_set_before(self.constraints_expr)
        self._assert_not_set_before(self.constraints_bounds)
        if D is not None and not isinstance(D, tuple):
            D = D, D
        ret = copy(self)
        ret.constraints_expr = g
        ret.constraints_bounds = D
        return ret

    def subject_to_penalty(
        self,
        g: Union[cs.SX, cs.MX],
        D: Optional[Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]] = None,
    ) -> "MinimizationProblemDescription":
        """
        Add general constraints :math:`g(x) \\in D`, handled using a quadratic
        penalty method.
        """
        self._assert_not_set_before(self.penalty_constraints_expr)
        self._assert_not_set_before(self.penalty_constraints_bounds)
        if D is not None and not isinstance(D, tuple):
            D = D, D
        ret = copy(self)
        ret.penalty_constraints_expr = g
        ret.penalty_constraints_bounds = D
        return ret

    def with_l1_regularizer(
        self, λ: Union[float, np.ndarray]
    ) -> "MinimizationProblemDescription":
        """
        Add an :math:`\\ell_1`-regularization term :math:`\\|\\lambda x\\|_1`
        to the objective.
        """
        self._assert_not_set_before(self.regularizer)
        ret = copy(self)
        ret.regularizer = λ
        return ret

    def with_param(
        self, p: Union[cs.SX, cs.MX], value: np.ndarray = None
    ) -> "MinimizationProblemDescription":
        """
        Make the problem depend on a symbolic parameter, with an optional
        default value. The value can be changed after the problem has been
        loaded, as wel as in between solves.
        """
        self._assert_not_set_before(self.parameter)
        ret = copy(self)
        ret.parameter = p
        if value is not None:
            ret.parameter_value = value
        return ret

    def with_param_value(self, value: np.ndarray) -> "MinimizationProblemDescription":
        """
        Explicitly change the parameter value for the parameter added by
        :py:func:`with_param`.
        """
        if self.parameter is None:
            raise RuntimeError("problem has no parameters")
        ret = copy(self)
        ret.parameter_value = value
        return ret

    def with_name(self, name: str) -> "MinimizationProblemDescription":
        """
        Set the name of the problem. Must be a valid file name, and for compiled
        problems, it should also be a valid C identifier, therefore it is
        recommended to stick to letters, numbers and underscores.
        """
        ret = copy(self)
        ret.name = name
        return ret

    def _convert_parts(self) -> dict:
        """
        Build a dictionary with all necessary components to build an alpaqa
        problem, i.e. the functions f and g, the sets C and D, etc.
        """
        # Function arguments (variables and parameters)
        x, param = self.variable, self.parameter
        args = [cs.vec(x), cs.vec(param)] if param is not None else [cs.vec(x)]
        # Objective and constraints functions
        f = [self.objective_expr]
        g_qpm = self.penalty_constraints_expr
        g_alm = self.constraints_expr
        g = cs.vertcat(*(g for g in (g_qpm, g_alm) if g is not None))
        g = [] if g.shape == (0, 0) else [g]
        # Problem dimensions
        n = cs.vec(x).shape[0]
        p = cs.vec(param).shape[0] if param is not None else 0
        m_qpm = g_qpm.shape[0] if g_qpm is not None else 0
        m_alm = g_alm.shape[0] if g_alm is not None else 0
        # Bound constraints
        C = self.bounds
        if C is None:
            C = (-np.inf * np.ones(n), +np.inf * np.ones(n))
        # General quadratic penalty method constraint set
        D_qpm = self.penalty_constraints_bounds
        if D_qpm is None:
            D_qpm = (-np.inf * np.ones(m_qpm), +np.inf * np.zeros(m_qpm))
        # General augmented Lagrangian method constraint set
        D_alm = self.constraints_bounds
        if D_alm is None:
            D_alm = (-np.inf * np.ones(m_alm), +np.inf * np.zeros(m_alm))
        num_param = self.parameter_value
        if num_param is None:
            num_param = np.nan * np.ones(p)
        λ = self.regularizer
        return dict(
            f=cs.Function("f", args, f),
            g=cs.Function("g", args, g) if g else None,
            C=C,
            D=np.hstack((D_qpm, D_alm)),
            param=num_param,
            l1_reg=λ,
            penalty_alm_split=m_qpm,
            name=self.name,
        )

    def compile(self, **kwargs) -> CasADiProblem:
        """
        Generate, compile and load the problem.

        A C compiler is required (e.g. GCC or Clang on Linux, Xcode on macOS, or
        Visual Studio on Windows).
        If no compiler is available, you could use the
        :py:meth:`alpaqa.MinimizationProblemDescription.build`
        method instead.

        :param \\**kwargs:
            Arguments passed to :py:func:`alpaqa.casadi_loader.generate_and_compile_casadi_problem`.

        :Keyword Arguments:
            * **second_order**: ``str`` --
              Whether to generate functions for evaluating second-order
              derivatives:

              * ``'no'``: only first-order derivatives (default).
              * ``'full'``: Hessians and Hessian-vector products of the
                Lagrangian and the augmented Lagrangian.
              * ``'prod'``: Hessian-vector products of the
                Lagrangian and the augmented Lagrangian.
              * ``'L'``: Hessian of the Lagrangian.
              * ``'L_prod'``: Hessian-vector product of the Lagrangian.
              * ``'psi'``: Hessian of the augmented Lagrangian.
              * ``'psi_prod'``: Hessian-vector product of the augmented
                Lagrangian.
            * **name**: ``str`` --
              Optional string description of the problem (used for filenames).
            * **sym**: ``Callable`` --
              Symbolic variable constructor, usually either ``cs.SX.sym``
              (default) or ``cs.MX.sym``.
              ``SX`` expands the expressions and generally results in better
              run-time performance, while ``MX`` usually has faster compile
              times.

        """

        return generate_and_compile_casadi_problem(
            **self._convert_parts(),
            **kwargs,
        )

    def build(self, **kwargs) -> CasADiProblem:
        """
        Finalize the problem formulation and return a problem type that can be
        used by the solvers.

        This method is usually not recommended: the
        :py:meth:`alpaqa.MinimizationProblemDescription.compile`
        method is preferred because it pre-compiles the problem for better
        performance.

        :Keyword Arguments:
            * **second_order**: ``str`` --
              Whether to generate functions for evaluating second-order
              derivatives:

              * ``'no'``: only first-order derivatives (default).
              * ``'full'``: Hessians and Hessian-vector products of the
                Lagrangian and the augmented Lagrangian.
              * ``'prod'``: Hessian-vector products of the
                Lagrangian and the augmented Lagrangian.
              * ``'L'``: Hessian of the Lagrangian.
              * ``'L_prod'``: Hessian-vector product of the Lagrangian.
              * ``'psi'``: Hessian of the augmented Lagrangian.
              * ``'psi_prod'``: Hessian-vector product of the augmented
                Lagrangian.
            * **sym**: ``Callable`` --
              Symbolic variable constructor, usually either ``cs.SX.sym``
              (default) or ``cs.MX.sym``.
              ``SX`` expands the expressions and generally results in better
              run-time performance.

        """
        parts = self._convert_parts()
        functions = _prepare_casadi_problem(parts["f"], parts["g"], **kwargs)
        ser_funcs = {k: v.serialize() for k, v in functions.items()}
        problem: CasADiProblem = deserialize_casadi_problem(ser_funcs)
        C, D = problem.variable_bounds, problem.general_bounds
        C.lower, C.upper = parts["C"]
        D.lower, D.upper = parts["D"]
        if parts["param"] is not None:
            problem.param = parts["param"]
        if parts["l1_reg"] is not None:
            problem.l1_reg = parts["l1_reg"]
        problem.penalty_alm_split = parts["penalty_alm_split"]
        return problem


def minimize(
    f: Union[cs.SX, cs.MX], x: Union[cs.SX, cs.MX]
) -> MinimizationProblemDescription:
    """
    Formulate a minimization problem with objective function :math:`f(x)` and
    unknown variables :math:`x`.
    """
    return MinimizationProblemDescription(objective_expr=f, variable=x)


__all__ = [
    "MinimizationProblemDescription",
    "minimize",
]



import alpaqa as pa
import numpy as np
from pprint import pprint
import pytest


@pytest.mark.skipif(
    not pa.with_casadi or not pa.with_ipopt, reason="requires CasADi and Ipopt"
)
def test_ipopt_pyapi_compile():
    import casadi as cs

    n = 2
    m = 2
    x = cs.SX.sym("x", n)

    Q = np.array([[1.5, 0.5], [0.5, 1.5]])
    f = 0.5 * x.T @ Q @ x
    g = x
    D = [-np.inf, 0.5], [+np.inf, +np.inf]
    p = pa.minimize(f, x).subject_to(g, D).compile(second_order="L")
    print(p)
    solver = pa.IpoptSolver({})
    cnt = pa.problem_with_counters(p)
    x0 = np.array([3, 3])
    y0 = np.zeros((m,))

    x, y, stats = solver(cnt.problem, x=x0, y=y0)

    print()
    print(cnt.evaluations)
    print(stats["status"])
    print("x", x)
    print("y", y)
    pprint(stats)
    assert stats["status"] == pa.IpoptSolverReturn.SUCCESS
    assert np.linalg.norm(x - [-1 / 6, 0.5]) < 1e-5
    assert np.linalg.norm(y - [0, -2 / 3]) < 1e-5
